import typing as t

class HostPathStr(t.NamedTuple):
    host: str
    path: str

def parse_hostpath(s: str) -> HostPathStr:
    split = s.split(":")
    if len(split) != 2:
        raise ValueError(f"HostPath \"{s}\" must be of format host:path")

    if not split[1].startswith("/"):
        raise ValueError(f"HostPath \"{s}\" must be an absolute path.")

    return HostPathStr(*split)


import logging
from argparse import ArgumentParser

from yog.host.manage import apply_necronomicon
from yog.logging_utils import setup


def main():

    args = ArgumentParser()
    args.add_argument("host")
    args.add_argument("--root-dir", default="./")
    args.add_argument("--debug", action="store_true")

    opts = args.parse_args()
    log = setup("yog", logging.DEBUG if opts.debug else logging.INFO)
    log.debug(f"Invoked with: {opts}")
    apply_necronomicon(opts.host, opts.root_dir)


import logging
import os
import typing as t
import re
from argparse import ArgumentParser
import json

import yaml
from docker.models.images import Image

from yog.git_utils import require_clean_work_tree
from subprocess import check_output
import docker
from dataclasses import dataclass

from yog.logging_utils import setup
from yog.ssh_utils import ScopedProxiedRemoteSSHTunnel

log = logging.getLogger(__name__)


@dataclass
class PushTarget:
    dockerfile_path: str
    registry_url: str = None
    tunnel_host: str = None
    tunnel_port: int = 5000
    context_path: str = None
    tags: t.List[str] = None
    repository: str = None

    def validate(self):
        if not self.is_private_registry() and not bool(self.tags):
            raise ValueError("You need to specify explicit tags to push if you're pushing to docker hub.")

    def is_private_registry(self):
        return bool(self.registry_url)


def _load_targets() -> t.Dict[str, PushTarget]:
    with open("./yog-repo.conf", "r") as fin:
        obj = yaml.full_load(fin)
    targets = {target_name: PushTarget(**target_dict) for target_name, target_dict in obj.items()}
    return targets

def ls():
    log.info("Available targets:")
    target: PushTarget
    for tname, target in _load_targets().items():
        log.info(f"{tname}: {target.registry_url}/{target.repository}")

def push(target: str, check_unclean_work_tree: bool):
    targets = _load_targets()

    push_target: PushTarget = targets[target]

    push_target.validate()

    if push_target.context_path:
        os.chdir(push_target.context_path)

    if check_unclean_work_tree and not require_clean_work_tree():
        raise RuntimeError("Your work tree is not clean.")

    revision = check_output(["git", "rev-parse", "HEAD"]).decode("utf-8").strip()

    if push_target.tunnel_host:
        log.info(f"Establishing tunnel to {push_target.tunnel_host}:{push_target.tunnel_port}")
        tunnel = ScopedProxiedRemoteSSHTunnel(push_target.tunnel_host, push_target.tunnel_port)
        tunnel_port = tunnel.connect()
    else:
        tunnel = None
        tunnel_port = None

    try:
        if push_target.tunnel_host:
            tag = f"localhost:{tunnel_port}/{push_target.repository}:{revision}"
        elif push_target.registry_url:
            tag = f"{push_target.registry_url}/{push_target.repository}:{revision}"
        else:
            tag = f"{push_target.repository}:{revision}"
        log.info(f"tag: {tag}")

        client: docker.DockerClient = docker.DockerClient(base_url='unix://var/run/docker.sock')
        log.info(f"Dockerfile: {push_target.dockerfile_path}")

        labels: t.Dict[str, str] = dict()
        labels["cafe.josh.yog.git-head-sha"] = revision

        log.info("Building...")
        image: Image
        image, logs = client.images.build(path=".", dockerfile=push_target.dockerfile_path, rm=True, tag=tag, labels=labels)

        log.info("Tagging...")
        if push_target.tags:
            for extra_tag in push_target.tags:
                image.tag(push_target.repository, extra_tag)

        log.info("Pushing...")
        if push_target.is_private_registry():
            log_lines = client.images.push(tag).splitlines()
        else:
            log_lines = []
            for extra_tag in push_target.tags:
                print(extra_tag)
                log_lines.extend(client.images.push(push_target.repository, extra_tag).splitlines())

        for line_raw in log_lines:
            if "sha256" in line_raw:
                log.info(line_raw)
            line = json.loads(line_raw)
            if "error" in line:
                raise RuntimeError(str(line))
        log.info("Done.")

    finally:
        if tunnel:
            tunnel.disconnect()


def main():
    setup("yog")
    args = ArgumentParser()

    args.add_argument("--workdir", default=None)

    subparsers = args.add_subparsers(help="Subcommand arguments.", dest="subcommand")
    push_parser = subparsers.add_parser("push")
    push_parser.add_argument("target")
    push_parser.add_argument("--no-git-check", action="store_true")

    list_parser = subparsers.add_parser("ls")

    prune_parser = subparsers.add_parser("prune")

    opts = args.parse_args()
    log.debug(f"Invoked with: {opts}")

    if opts.workdir:
        log.info(f"chdir: {opts.workdir}")
        os.chdir(opts.workdir)

    if opts.subcommand == "push":
        push(opts.target, not opts.no_git_check)
    elif opts.subcommand == "ls":
        ls()
    else:
        log.error(f"Invalid command: {opts.subcommand}")








import logging
import os.path
from os.path import dirname, join, isdir, isfile, basename

import docker
from docker.models.containers import Container
from paramiko import SSHClient, SSHException
from paramiko.ssh_exception import NoValidConnectionsError
from yog.host.pki_model import load_cas

from yog.host import necronomicon
from yog.host.docker_attrs import build_run_kwargs_dict, diff_container
from yog.host.necronomicon import Necronomicon, File
import yog.host.pki as pki
from yog.ssh_utils import check_call, ScopedProxiedRemoteSSHTunnel, compare_local_and_remote

log = logging.getLogger(__name__)


def apply_necronomicon(host: str, root_dir):
    ssh = SSHClient()
    ssh.load_system_host_keys()
    try:
        log.info(f"[{host}]")
        ssh.connect(host)
        apply_necronomicon_for_host(host, ssh, root_dir)
    except RuntimeError as e:
        log.error(f"{host} error: {e.__class__.__name__}: {str(e)}")
    except SSHException as e:
        log.error(f"{host} error: {e.__class__.__name__}: {str(e)}")
    except NoValidConnectionsError as e:
        log.error(f"{host} error: {e.__class__.__name__}: {str(e)}")
    finally:
        ssh.close()


def load_necronomicons_for_host(host: str, root_dir):
    necronomicon_paths = []
    cur = join(root_dir, "domains")
    if isfile(join(cur, "_.yml")):
        necronomicon_paths.append(join(cur, "_.yml"))

    for part in reversed(host.split(".")):
        if not isdir(cur):
            break
        if isfile(join(cur, f"{part}.yml")):
            necronomicon_paths.append(join(cur, f"{part}.yml"))
        cur = join(cur, part)

    return [necronomicon.loadfile(basename(p), p) for p in necronomicon_paths]


def apply_necronomicon_for_host(host: str, ssh: SSHClient, root_dir):
    necronomicons = load_necronomicons_for_host(host, root_dir)
    if not necronomicons:
        raise RuntimeError(f"No necronomicons found for {host}")

    necronomicons = [n.inflate(host, ssh) for n in necronomicons]

    for n in necronomicons:
        if n.files.files:
            apply_files_section(host, n, ssh, root_dir)
        if n.docker.containers:
            apply_docker_section(host, n)
        if n.cron.crons:
            apply_cron_section(host, n, ssh)
        if n.pki.certs:
            pki.apply_pki_section(host, n, ssh, root_dir)


def apply_cron_section(host: str, n: Necronomicon, ssh: SSHClient):
    cronfile_lines = []
    line1_length = max([len(c.expr) for c in n.cron.crons])
    line2_length = max([len(c.user) for c in n.cron.crons])
    for cron in n.cron.crons:
        cronfile_lines.append(f"{cron.expr.ljust(line1_length, ' ')}\t{cron.user.ljust(line2_length, ' ')}\t{cron.command}")
    cronfile_body = "\n".join(cronfile_lines + [""])
    ok, expected, found = compare_local_and_remote(bytes(cronfile_body, "utf-8"), "/etc/cron.d/yog", ssh)
    if not ok:
        log.info(f"[{host}][cron]: stale")
        log.info(f"Writing /etc/cron.d/yog version {expected[:10]}")
        check_call(ssh, "sudo bash -c 'cat - > /etc/cron.d/yog'", send_stdin=cronfile_body)
    else:
        log.info(f"[{host}][cron]: OK")


def apply_docker_section(host: str, n: Necronomicon):
    log.debug(f"Docker: {n.ident}")
    tunnels = []
    for tunnel_def in n.tunnels.tunnels:
        log.debug(f"Setting up tunnel {tunnel_def}")
        tunnel = ScopedProxiedRemoteSSHTunnel(
            host,
            tunnel_def.target_port,
            tunnel_def.host,
            "remote",
            force_random_port=tunnel_def.local_port)
        tunnels.append(tunnel)

    try:
        for tun in tunnels:
            tun.connect()
        with ScopedProxiedRemoteSSHTunnel(host, 4243) as tport:
            log.debug(f"Connecting docker client.... tcp://localhost:{tport}")
            client = docker.DockerClient(base_url=f"tcp://localhost:{tport}", version="auto")
            log.debug("Docker connected.")

            for desired_container in n.docker.containers:
                log.debug(desired_container)

                log.debug(f"PULL: {desired_container.image}@{desired_container.fingerprint}")

                cur = client.containers.list(all=True, filters={"name": desired_container.name})
                c: Container = cur[0] if len(cur) > 0 else None

                if c:
                    log.debug(f"Existing container {c.id} is {c.status}")
                    diffs = diff_container(c, desired_container)
                    if not diffs:
                        log.debug(f"{c.id} is image {c.image.id} which matches our target.")
                        log.info(f"[{host}][docker]: OK {desired_container.name}@{desired_container.fingerprint[7:13]}")
                        log.debug("Existing container matches our desired state. No need to kill it.")
                        continue

                    log.debug(f"[{host}][docker][{c.name}]: diffs:")
                    for diff in diffs:
                        log.debug(f"[{host}][docker][{c.name}]: {diff[0]}")
                        log.debug(f"[{host}][docker][{c.name}]: desired {diff[1]}")
                        log.debug(f"[{host}][docker][{c.name}]: found {diff[2]}")

                    if c.status in ["running", "restarting"]:
                        log.debug(f"STOP {c.name}:{c.id}")
                        c.stop()
                    log.debug(f"RM: {c.name}:{c.id}")
                    c.remove()

                log.debug(f"RUN: {desired_container.image}@{desired_container.fingerprint}")
                log.info(f"[{host}][docker]: run {desired_container.name}")
                client.containers.run(**build_run_kwargs_dict(desired_container))

    finally:
        for tun in tunnels:
            try:
                tun.disconnect()
            except RuntimeError as e:
                log.warning("Error while disconnecting tunnel", exc_info=e)


def _load_file_content(f: File, root_dir: str) -> bytes:
    if f.src.startswith("ca:"):
        ca_ident = f.src[len("ca:"):]
        matches = [ca for ca in load_cas(os.path.join(root_dir, "cas.yml")) if ca_ident == ca.ident]
        if not matches:
            raise ValueError(f"No such CA with ident {ca_ident}")
        ca = matches[0]
        ssh = SSHClient()
        ssh.load_system_host_keys()
        ssh.connect(ca.storage.host)
        try:
            trust = pki.KeyPairData.load(ssh, ca.storage.path)
        finally:
            ssh.close()

        return trust.raw_crt().body.encode("utf-8")


    with open(join(root_dir, "files", f.src)) as fin:
        return fin.read().encode("utf-8")


def apply_files_section(host: str, n: Necronomicon, ssh: SSHClient, root_dir):
    log.debug(f"Files: {n.ident}")
    hupcmds = set()
    for f in n.files.files:
        content = _load_file_content(f, root_dir)
        ok, _, _ = compare_local_and_remote(content, f.dest, ssh, f.root)
        if ok:
            log.info(f"[{host}][files]: OK [{f.src}]")
        else:
            log.info(f"[{host}][files]: stale {f.src} -> {f.dest}")
            if f.hupcmd:
                hupcmds.add(f.hupcmd)

            if f.root:
                cmd_prefix = "sudo "
            else:
                cmd_prefix = ""

            check_call(ssh, f"{cmd_prefix}mkdir -p \"{dirname(f.dest)}\"")
            check_call(ssh, f"{cmd_prefix}bash -c 'cat - > \"{f.dest}\"'", send_stdin=content)

    for c in hupcmds:
        log.info(f"[{host}][files][hup]: {c}")
        check_call(ssh, c)






import logging
import os
import sys
from getpass import getuser
from os.path import join, exists

log = logging.getLogger(__name__)


def setup(app: str, stdout_level: int = logging.INFO):
    ret = logging.getLogger(f"yog")
    ret.setLevel(logging.DEBUG)

    if getuser() == "root":
        log_dir = join("/var", "log", app)
    else:
        log_dir = join(os.environ["HOME"], ".cache", app)

    if not exists(log_dir):
        os.makedirs(log_dir, exist_ok=True)

    logfile = logging.FileHandler(join(log_dir, "log.txt"))
    logfile.setLevel(logging.DEBUG)
    logfile.setFormatter(logging.Formatter("[%(asctime)s] [%(levelname)s]: %(message)s [%(pathname)s] [L%(lineno)d]"))
    ret.addHandler(logfile)

    stdout = logging.StreamHandler(sys.stdout)
    stdout.setLevel(stdout_level)
    stdout.setFormatter(logging.Formatter("%(message)s"))
    ret.addHandler(stdout)

    ret.debug("Logging configured.")

    return ret




import logging
from argparse import ArgumentParser

from yog.logging_utils import setup
from yog.host.pki import apply_cas


def ca_main():
    args = ArgumentParser()
    args.add_argument("--ident", default=None)
    args.add_argument("--root-dir", default="./")
    args.add_argument("--debug", action="store_true")

    opts = args.parse_args()
    log = setup("yog-ca", logging.DEBUG if opts.debug else logging.INFO)

    apply_cas(opts.ident, opts.root_dir)




import os
import re
import typing as t
import yaml

from yog.model_utils import HostPathStr, parse_hostpath


class CAEntry(t.NamedTuple):
    ident: str
    storage: HostPathStr
    validity_period: str


def parse_validity_period(s: str) -> int:
    if s.endswith("y"):
        return 365 * int(s[:-1])
    elif s.endswith("d"):
        return int(s[:-1])
    else:
        raise ValueError(f"Invalid validity period: needs to end in y or d: {s}")


def load_caentry(raw: t.Any) -> CAEntry:
    return CAEntry(raw["ident"], parse_hostpath(raw["storage"]), str(raw["validity_period"]))


def load_cas(path: str) -> t.List[CAEntry]:
    with open(path, "r") as fin:
        raw = yaml.safe_load(fin.read())
    return [load_caentry(re) for re in raw]




import datetime
import logging
import os
import typing as t
import uuid

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePrivateKey
from cryptography.hazmat.primitives.serialization import load_pem_private_key
from cryptography.x509 import Certificate, GeneralName
from cryptography.x509.oid import NameOID
from paramiko.client import SSHClient

from yog.host.necronomicon import CertEntry, Necronomicon
from yog.host.os_utils import Perms, Owner, RWXBits
from yog.host.pki_model import CAEntry, load_cas, parse_validity_period
from yog.ssh_utils import mkdirp, cat, exists, put, check_call, check_output, stat

log = logging.getLogger(__name__)

class KeyMaterial(t.NamedTuple):
    fname: str
    mattype: str
    body: str
    owner: t.Optional[Owner]
    perms: t.Optional[Perms]


class KeyPairDataItem(t.NamedTuple):
    field_name: str
    fname: str
    material_type: str

KEY_PAIR_DATA_ITEMS = [
    KeyPairDataItem("private_openssl", "key.pem.openssl", "private"),
    KeyPairDataItem("private_ssh", "key.ssh", "private"),
    KeyPairDataItem("public_pkcs", "key.pem.pkcs1.public", "public"),
    KeyPairDataItem("public_ssh", "key.ssh.public", "public"),
    KeyPairDataItem("certificate", "key.crt", "cert"),
]


class KeyPairData(t.NamedTuple):
    serial: uuid.UUID
    issuer_serial: uuid.UUID
    private_ssh: KeyMaterial
    private_openssl: KeyMaterial

    public_pkcs: KeyMaterial
    public_ssh: KeyMaterial

    certificate: KeyMaterial

    def materials(self) -> t.List[KeyMaterial]:
        return [self.private_ssh, self.private_openssl, self.public_ssh, self.public_pkcs, self.certificate]

    def raw_crt(self) -> KeyMaterial:
        return self.certificate

    def crt(self) -> Certificate:
        return x509.load_pem_x509_certificate(self.raw_crt().body.encode("utf-8"))

    def private(self) -> EllipticCurvePrivateKey:
        return load_pem_private_key(self.private_openssl.body.encode("utf-8"), None)

    def cert_names(self) -> t.List[str]:
        e: x509.SubjectAlternativeName = self.crt().extensions.get_extension_for_class(x509.SubjectAlternativeName).value
        return e.get_values_for_type(x509.DNSName)

    def issuer_cn(self) -> str:
        return self.crt().issuer.get_attributes_for_oid(NameOID.COMMON_NAME)[0].value

    def write(self, ssh: SSHClient, path: str, owner: t.Optional[Owner]=None, perms: t.Optional[Perms]=None):
        mkdirp(ssh, path, "root")
        for km in self.materials():
            put(ssh, os.path.join(path, km.fname), km.body,
                user=owner.user if owner else None,
                group=owner.group if owner else None,
                mode=("600" if km.mattype == "private" else perms.to_chmod_expr() if perms else "644"))
        put(ssh, os.path.join(path, "issuer_serial.txt"), str(self.issuer_serial), "root")
        put(ssh, os.path.join(path, "serial.txt"), str(self.serial), "root")

    @staticmethod
    def load(ssh: SSHClient, path: str) -> 'KeyPairData':
        serial_path = os.path.join(path, "serial.txt")
        issuer_serial_path = os.path.join(path, "issuer_serial.txt")
        if (not exists(ssh, issuer_serial_path)) or (not exists(ssh, serial_path)):
            raise ValueError(f"Unable to load {path} (missing serial file)")

        serial = uuid.UUID(cat(ssh, serial_path))
        issuer_serial = uuid.UUID(cat(ssh, issuer_serial_path))

        args = {}
        for i in KEY_PAIR_DATA_ITEMS:
            if not exists(ssh, os.path.join(path, i.fname)):
                raise ValueError(f"Unable to load {path} (missing key data)")
            args[i.field_name] = KeyMaterial(
                i.fname,
                i.material_type,
                cat(ssh, os.path.join(path, i.fname), i.material_type == "private"),
                *stat(ssh, os.path.join(path, i.fname))
            )

        return KeyPairData(serial, issuer_serial, **args) # TODO these Nones


def _gen_ca(ca: CAEntry):
    private_key = ec.generate_private_key(
        curve = ec.SECP384R1(),
        backend=default_backend()
    )

    public_key = private_key.public_key()
    builder = x509.CertificateBuilder()
    builder = builder.subject_name(x509.Name([
        x509.NameAttribute(NameOID.COMMON_NAME, ca.ident),
    ]))
    builder = builder.issuer_name(x509.Name([
        x509.NameAttribute(NameOID.COMMON_NAME, ca.ident),
    ]))
    builder = builder.not_valid_before(datetime.datetime.utcnow())
    builder = builder.not_valid_after(datetime.datetime.utcnow()+datetime.timedelta(days=parse_validity_period(ca.validity_period)))
    serial_no = uuid.uuid4()
    builder = builder.serial_number(int(serial_no))
    builder = builder.public_key(public_key)
    builder = builder.add_extension(
        x509.BasicConstraints(ca=True, path_length=None), critical=True,
    )
    certificate = builder.sign(
        private_key=private_key, algorithm=hashes.SHA256(),
        backend=default_backend()
    )

    material = KeyPairData(
        serial_no,
        serial_no,
        KeyMaterial("key.ssh", "private", private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.OpenSSH,
            encryption_algorithm=serialization.NoEncryption(),
            ).decode("utf-8"), None, None),
        KeyMaterial("key.pem.openssl", "private", private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=serialization.NoEncryption(),
        ).decode("utf-8"), None, None),
        KeyMaterial("key.pem.pkcs1.public", "public",
            public_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo,
            ).decode("utf-8"), None, None),
        KeyMaterial("key.ssh.public", "public",
            public_key.public_bytes(
            encoding=serialization.Encoding.OpenSSH,
            format=serialization.PublicFormat.OpenSSH,
            ).decode("utf-8"), None, None),
        KeyMaterial("key.crt", "cert",
            certificate.public_bytes(
            encoding=serialization.Encoding.PEM,
            ).decode("utf-8"), None, None),
    )

    return material


def _gen_cert(ce: CertEntry, ca_data: KeyPairData, ca: CAEntry):
    private_key = ec.generate_private_key(
        curve = ec.SECP384R1(),
        backend=default_backend()
    )

    public_key = private_key.public_key()
    builder = x509.CertificateBuilder()
    builder = builder.subject_name(x509.Name([
        x509.NameAttribute(NameOID.COMMON_NAME, ce.names[0]),
    ]))
    builder = builder.issuer_name(x509.Name([
        x509.NameAttribute(NameOID.COMMON_NAME, ca.ident),
    ]))
    builder = builder.not_valid_before(datetime.datetime.utcnow())
    builder = builder.not_valid_after(datetime.datetime.utcnow()+datetime.timedelta(days=parse_validity_period(ce.validity_period)))
    serial_no = uuid.uuid4()
    builder = builder.serial_number(int(serial_no))
    builder = builder.public_key(public_key)
    builder = builder.add_extension(
        x509.SubjectAlternativeName([x509.DNSName(n) for n in ce.names]), critical=False
    )
    certificate = builder.sign(
        private_key=ca_data.private(), algorithm=hashes.SHA256(),
        backend=default_backend()
    )

    material = KeyPairData(
        serial_no,
        ca_data.issuer_serial,
        KeyMaterial("key.ssh", "private", private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.OpenSSH,
            encryption_algorithm=serialization.NoEncryption(),
        ).decode("utf-8"), None, None),
        KeyMaterial("key.pem.openssl", "private", private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=serialization.NoEncryption(),
            ).decode("utf-8"), None, None),
        KeyMaterial("key.pem.pkcs1.public", "public",
            public_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo,
            ).decode("utf-8"), None, None),
        KeyMaterial("key.ssh.public", "public",
            public_key.public_bytes(
            encoding=serialization.Encoding.OpenSSH,
            format=serialization.PublicFormat.OpenSSH,
            ).decode("utf-8"), None, None),
        KeyMaterial("key.crt", "cert",
            certificate.public_bytes(
            encoding=serialization.Encoding.PEM,
            ).decode("utf-8"), None, None),
    )

    return material




def apply_cas(ident: t.Optional[str], root_dir: str):
    cas = load_cas(os.path.join(root_dir, "cas.yml"))
    if ident:
        cas = [ca for ca in cas if ca.ident == ident]

    for ca in cas:
        _apply_ca(ca)

def _apply_ca(ca: CAEntry):
    ssh = SSHClient()
    ssh.load_system_host_keys()

    ssh.connect(ca.storage.host)
    try:
        _provision_hier(ssh, ca)
        try:
            cadata = KeyPairData.load(ssh, ca.storage.path)
        except ValueError:
            cadata = None
        if not cadata:
            _provision_ca(ssh, ca)
        else:
            log.info("CA is OK")
    finally:
        ssh.close()

def _provision_hier(ssh: SSHClient, ca: CAEntry):
    mkdirp(ssh, ca.storage.path, "root")


def _provision_ca(ssh: SSHClient, ca: CAEntry):
    log.info("Generating new CA...")
    cadata = _gen_ca(ca)
    cadata.write(ssh, ca.storage.path)


def apply_pki_section(host: str, n: Necronomicon, ssh: SSHClient, root_dir):
    cas = load_cas(os.path.join(root_dir, "cas.yml"))

    hupcmds = set()

    for ce in n.pki.certs:
        generate = False

        root = [ca for ca in cas if ca.ident == ce.authority]
        if not root:
            raise ValueError(f"No such CA: {ce.authority}")
        ca = root[0]

        ssh_ca = SSHClient()
        ssh_ca.load_system_host_keys()
        ssh_ca.connect(ca.storage.host)
        try:
            ca_data = KeyPairData.load(ssh_ca, ca.storage.path)
        finally:
            ssh_ca.close()

        try:
            cur_trust = KeyPairData.load(ssh, ce.storage)
            cert: Certificate = cur_trust.crt()
            expiry = cert.not_valid_after
            if (expiry - datetime.timedelta(days=parse_validity_period(ce.refresh_at_period))) <= datetime.datetime.utcnow():
                generate = True
                log.debug("Expiry too soon")
            elif set(ce.names) != set(cur_trust.cert_names()):
                generate = True
                log.debug("Set of names != cert names")
            elif cur_trust.issuer_cn() != ce.authority:
                generate = True
                log.debug("Issuer CN != authority ident")
            elif expiry > (datetime.datetime.utcnow() + datetime.timedelta(days=parse_validity_period(ce.validity_period))):
                generate = True
                log.debug("expiry too far out")
            elif cert.subject.get_attributes_for_oid(NameOID.COMMON_NAME)[0].value != ce.names[0]:
                generate = True
                log.debug("CN != cert CN")
            elif ca_data.serial != cur_trust.issuer_serial:
                generate = True
                log.debug("CA serial != cert issuer serial")
            elif ce.chown and any(ce.chown != m.owner for m in cur_trust.materials() if m.mattype != "private"):
                generate = True
                log.debug("chown != current ownership")
            elif ce.chmod and any(Perms(*RWXBits.from_stat(ce.chmod)) != m.perms for m in cur_trust.materials() if m.mattype != "private"):
                generate = True
                log.debug("chmod != current perms")
        except ValueError:
            generate = True
            log.debug("no cert found")

        if not generate:
            log.info(f"[{host}][pki]: OK [{ce.storage}]")
            continue

        log.info(f"[{host}][pki]: stale [{ce.storage}]")

        trust_new = _gen_cert(ce, ca_data, ca)
        trust_new.write(ssh, ce.storage, None if not ce.chown else Owner.from_str(ce.chown), None if not ce.chmod else Perms(*RWXBits.from_stat(ce.chmod)))

        if ce.hupcmd:
            hupcmds.add(ce.hupcmd)

    for cmd in hupcmds:
        log.info(f"[{host}][pki][hup]: {cmd}")
        check_call(ssh, cmd)



import typing as t


class RWXBits(t.NamedTuple):
    read: bool
    write: bool
    execute: bool

    @staticmethod
    def from_str(str_repr: str) -> 'RWXBits':
        return RWXBits("r" in str_repr, "w" in str_repr, "x" in str_repr)

    @staticmethod
    def from_stat(stat_str: str) -> t.Tuple['RWXBits', 'RWXBits', 'RWXBits']:
        return (
            RWXBits.from_str(stat_str[0:3]),
            RWXBits.from_str(stat_str[3:6]),
            RWXBits.from_str(stat_str[6:9]),
        )

    def to_octal_digit(self) -> str:
        sum = 0
        if self.read:
            sum += 4
        if self.write:
            sum += 2
        if self.execute:
            sum += 1

        return str(sum)


class Owner(t.NamedTuple):
    user: str
    group: str

    @staticmethod
    def from_str(s: str) -> 'Owner':
        return Owner(*s.split(":", 1))


class Perms(t.NamedTuple):
    u: RWXBits
    g: RWXBits
    o: RWXBits

    def to_chmod_expr(self) -> str:
        return "".join(p.to_octal_digit() for p in [self.u, self.g, self.o])



import getpass
import typing as t
import logging
import os
import re
import subprocess
from hashlib import sha512
from random import randrange
from re import Pattern
from tempfile import NamedTemporaryFile
from time import sleep
from typing import Tuple, List, Optional, Set

from paramiko import SSHClient, PKey
import shlex

from yog.host.os_utils import Owner, RWXBits, Perms

log = logging.getLogger(__name__)


def check_call(ssh: SSHClient, command: str, assert_stderr_empty: bool = False, send_stdin: Optional[t.AnyStr] = None, ):
    stdin, stdout, stderr = ssh.exec_command(command)
    if send_stdin is not None:
        stdin.write(send_stdin)
    stdin.close()
    rc = stdout.channel.recv_exit_status()

    if rc != 0:
        print(f"'{command}' returned exit code {rc}")
        _dump_lines("stdout", stdout.readlines())
        _dump_lines("stderr", stderr.readlines())
        raise NonzeroExitCodeError(rc)

    if assert_stderr_empty:
        stderr_contents = stderr.readlines()
        if stderr_contents:
            print(f"'{command}' expected empty stderr but it wasn't empty.")
            _dump_lines("stderr", stderr.readlines())
            raise StdErrNotEmptyError(stderr_contents)


def check_output(ssh: SSHClient, command: str, send_stdin: Optional[str] = None) -> Tuple[List[str], List[str]]:
    stdin, stdout, stderr = ssh.exec_command(command)
    if send_stdin is not None:
        stdin.write(send_stdin)

    rc = stdout.channel.recv_exit_status()

    if rc != 0:
        _dump_lines("stdout", stdout.readlines())
        _dump_lines("stderr", stderr.readlines())
        raise NonzeroExitCodeError(rc)

    return stdout.readlines(), stderr.readlines()


def check_code(ssh: SSHClient,
               command: str,
               assert_stderr_empty: bool = False,
               send_stdin: Optional[str] = None,
               assert_stdout_empty: bool = False
               ) -> bool:
    stdin, stdout, stderr = ssh.exec_command(command)
    if send_stdin is not None:
        stdin.write(send_stdin)
    rc = stdout.channel.recv_exit_status()

    if rc != 0:
        return False

    if assert_stderr_empty:
        stderr_contents = stderr.readlines()
        if stderr_contents:
            return False

    if assert_stdout_empty:
        stdout_contents = stdout.readlines()
        if stdout_contents:
            return False

    return True


def check_stdout(ssh: SSHClient, command: str) -> t.List[str]:
    return check_output(ssh, command)[0]


def check_stderr(ssh: SSHClient, command: str) -> t.List[str]:
    return check_output(ssh, command)[1]


def _dump_lines(title: str, lines: List[str]):
    print(f"--------------{title}------------")
    for line in lines:
        print(line)
    print("--------------------------------")


class NonzeroExitCodeError(Exception):
    code: int

    def __init__(self, code: int):
        self.code = code


class StdErrNotEmptyError(Exception):
    contents: List[str]

    def __init__(self, contents: List[str]):
        self.contents = contents


def get_host_key(ssh: SSHClient) -> PKey:
    return ssh.get_transport().get_remote_server_key()


def get_pids_binding_port(bound_port: int) -> Set[int]:
    lsof_lines = str(subprocess.check_output(["lsof", "-i", f":{bound_port}"]), encoding="utf-8")
    lsof_lines = lsof_lines.splitlines()[1:]
    lsof_lines = [re.split(r"\s+", str(l)) for l in lsof_lines]
    pids_bound = {int(l[1]) for l in lsof_lines if l[4] == "IPv4"}
    return pids_bound


# always prints:
# debug1: Local forwarding listening on 127.0.0.1 port 44444.
# this means success:
# debug1: channel 0: new [port listener]

success_pattern: Pattern = re.compile(r"^(debug1: channel \d+: new \[port listener])|(debug1: remote forward success for: listen \d+, connect localhost:\d+)|(debug1: channel \d+: new port-listener \[port listener\] \(inactive timeout: \d+\))$")


def _render_cmd(rand_port, prefix, host, forward_flag, expr):
    return prefix + ["ssh", "-v", "-4", "-o", "ExitOnForwardFailure=yes",
                     "-N", forward_flag, expr.format(rand_port),
                     host]

# 2 bugs:
# * stderr will keep printing but i've stopped reading it. This will cause a deadlock eventually. fix by just moving to
#  a new thread and babysitting stderr there
# * when proxying, remote ssh process doesn't die when local ssh process dies. this keeps the port bound.

# crazy idea: fuck asyncio, fuck threads, etc. Popen(bash -c ssh 2>/tmp/namedtmpfile) and then fucking read
# the tempfile over and over till I find my shit.


class ScopedProxiedRemoteSSHTunnel:
    proxy_host: Optional[str]
    host: str
    port_forwarded: int
    forward_type: str

    proc: Optional[subprocess.Popen]

    def __init__(self, host: str,
                 port_forwarded: int,
                 proxy_host: Optional[str] = None,
                 forward_type: str = "local",
                 force_random_port: int = None):
        self.host = host
        self.port_forwarded = port_forwarded
        self.proxy_host = proxy_host
        self.forward_type = forward_type
        self.force_random_port = force_random_port

    def __enter__(self):
        return self.connect()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.disconnect()

    def connect(self) -> int:
        if self.forward_type == "local":
            forward_flag = "-L"
            forward_expr = f"{{}}:localhost:{self.port_forwarded}"
        elif self.forward_type == "remote":
            forward_flag = "-R"
            forward_expr = f"{self.port_forwarded}:localhost:{{}}"
        else:
            raise ValueError(f"Illegal value for forward_type: \"{self.forward_type}\" must be one of local, remote")

        if self.proxy_host is not None:
            cmd_prefix = ["ssh", "-t", "-t", self.proxy_host]
        else:
            cmd_prefix = []

        self.proc, rport = _setup_tunnel(cmd_prefix, self.host, forward_flag, forward_expr, self.force_random_port)
        return rport

    def disconnect(self):
        if self.proc is not None:
            self.proc.kill()
            log.debug(f"SSH tunnel ended.")


def _setup_tunnel(cmd_prefix, host, forward_flag, forward_expr, force_random_port=None) -> Tuple[subprocess.Popen, int]:
    tries = 25
    port = None
    proc = None
    while tries > 0 and proc is None:
        if force_random_port is None:
            port = randrange(1024, 65535 + 1)
        else:
            port = force_random_port
        cmd = _render_cmd(port, cmd_prefix, host, forward_flag, forward_expr)
        stderr_file = NamedTemporaryFile("w", delete=False)
        stdout_file = NamedTemporaryFile("w", delete=False)
        try:
            proc = _try_setup_tunnel(port, cmd, stdout_file, stderr_file)
        except RuntimeError as e:
            log.info(f"SSH tunnel open threw exception. Tries left: {tries}", exc_info=e)
            tries -= 1
        finally:
            os.remove(stderr_file.name)
            os.remove(stdout_file.name)

    if proc is None:
        log.error(f"Gave up. Couldn't open ssh tunnel.")
        raise RuntimeError(f"Couldn't open SSH tunnel.")

    if port is None:
        raise RuntimeError("port was none. this should never happen.")
    return proc, port


def _try_setup_tunnel(port: int, cmd: List[str], stdout_file: NamedTemporaryFile, stderr_file: NamedTemporaryFile) -> Optional[subprocess.Popen]:
    proc = subprocess.Popen(cmd, stdout=stdout_file, stderr=stderr_file, stdin=subprocess.PIPE)
    confirmed = False
    timeout = 10.0
    while not (confirmed or proc.returncode is not None) and timeout > 0:
        sleep(0.1)
        timeout -= 0.1
        for fobj in [stdout_file, stderr_file]:
            fobj.flush()
            with open(fobj.name, "r") as fin:
                for line in fin:
                    if success_pattern.match(line.strip()):
                        confirmed = True

    if not confirmed:
        outputs = []
        for fileobj in [stdout_file, stderr_file]:
            with open(fileobj.name, "r") as fin:
                outputs.append(fin.read())
        stderr = "\n".join(outputs)
        if "Address already in use" in stderr:
            log.info(f"Port {port} already in use.")
        else:
            log.error(f"SSH tunnel didn't stay open. Weird.")
            log.error("--------------ssh stderr-------------------")
            log.error(stderr)
            log.error("--------------end ssh stderr---------------")
        return None
    else:
        log.debug(f"SSH tunnel started.")
        return proc


def compare_local_and_remote(body: bytes, remote_path: str, ssh: SSHClient, root: bool = False):
    expected: str = sha512(body).hexdigest()
    remote_path = shlex.quote(remote_path)

    found: Optional[str]
    if check_code(ssh, f"{'sudo ' if root else ''}test -f \"{remote_path}\""):
        found: str = re.split(r"\s+", check_output(ssh, f"{'sudo ' if root else ''}sha512sum \"{remote_path}\"")[0][0], 1)[0]
    else:
        found = None

    return expected == found, expected, found

def cat(ssh: SSHClient, path: str, root: bool = False) -> str:
    log.debug(f"cat {path}")
    prefix = 'sudo ' if root else ''
    path = shlex.quote(path)
    cmd = prefix + f"cat {path}"
    return "".join(check_stdout(ssh, cmd))

def mkdirp(ssh: SSHClient, path: str, user: str = None):
    log.debug(f"mkdir -p {path}")
    path = shlex.quote(path)
    cmd = f"mkdir -p {path}"
    cmd = _as_user(cmd, user)
    check_call(ssh, cmd)


def stat(ssh: SSHClient, path: str) -> t.Tuple[Owner, Perms]:
    path = shlex.quote(path)
    # stat -c '%A %U:%G'
    raw = check_output(ssh, f"stat -c {shlex.quote('%A %U:%G')} {path}")[0][0]
    log.debug(f"stat: {path}: {raw}")
    perms, owner = raw.split(" ", 1)
    perms = perms[-9:]

    return Owner.from_str(owner), Perms(*RWXBits.from_stat(perms))


def _as_user(cmd: str, user: t.Optional[str]) -> str:
    user = shlex.quote(user)
    if user is None:
        return cmd
    elif user == "root":
        return "sudo " + cmd
    else:
        return f"sudo -u {user} {cmd}"


def exists(ssh: SSHClient, path: str) -> bool:
    log.debug(f"exists? {path}")
    path = shlex.quote(path)
    ret = check_code(ssh, f"test -e {path}")
    log.debug(f"exists? {path} = {ret}")
    return ret


def put(ssh: SSHClient, path: str, content: t.AnyStr, user: t.Optional[str] = None, group: t.Optional[str]=None, mode: t.Optional[str] = "644"):
    log.debug(f"put {path} {user}:{group} {mode}")
    mode = shlex.quote(mode)
    path = shlex.quote(path)

    install_opts = [f"-m {mode}"]
    if user:
        user = shlex.quote(user)
        install_opts.append(f"-o {user}")
    if group:
        group = shlex.quote(group)
        install_opts.append(f"-g {group}")
    check_call(ssh, _as_user(f"install {' '.join(install_opts)} /dev/null {path}", "root"))

    subcommand = shlex.quote(f"cat - > {path}")
    check_call(ssh, _as_user(f"bash -c {subcommand}", "root"), send_stdin=content)



import ipaddress
import re
import typing as t
from dataclasses import dataclass
from typing import Any

from docker.models.containers import Container
from docker.models.images import Image
from docker.types import LogConfig

from yog.host.necronomicon import DockerContainer, PortEntry, PortString


class DockerAttribute:

    def run_arg_name(self) -> str:
        raise NotImplemented()

    def from_necronomicon(self, dc: DockerContainer) -> 'DockerAttributeInstance':
        raise NotImplemented()

    def from_container(self, c: Container) -> 'DockerAttributeInstance':
        raise NotImplemented()


class DockerAttributeInstance:
    def to_run_arg(self) -> t.Any:
        raise NotImplemented()

    def is_satisfied_by(self, actual: 'DockerAttributeInstance'):
        raise NotImplemented()


def build_run_kwargs_dict(dc: DockerContainer) -> t.Dict[Any, Any]:
    return {a.run_arg_name(): a.from_necronomicon(dc).to_run_arg() for a in SUPPORTED_DOCKER_ATTRS}


class SimpleDockerAttributeInstance(DockerAttributeInstance):
    def __init__(self, o):
        self.o = o

    def to_run_arg(self) -> t.Any:
        return self.o

    def is_satisfied_by(self, other: 'SimpleDockerAttributeInstance'):
        return self.o == other.o

    def __repr__(self) -> str:
        return repr(self.o)


ImageName: t.TypeAlias = str
ImageDigest: t.TypeAlias = str


@dataclass
class DockerImageDefInstance(DockerAttributeInstance):
    repo_digests: t.List[t.Tuple[ImageName, ImageDigest]]

    def to_run_arg(self) -> t.Any:
        name, digest = self.repo_digests[0]
        return f"{name}@{digest}"

    def is_satisfied_by(self, other: 'DockerImageDefInstance'):
        return any(rd in other.repo_digests for rd in self.repo_digests)


class DockerImageDef(DockerAttribute):
    def run_arg_name(self) -> str:
        return "image"

    def from_necronomicon(self, dc: DockerContainer) -> 'DockerAttributeInstance':
        return DockerImageDefInstance([(dc.image, dc.fingerprint)])

    def from_container(self, c: Container) -> t.Any:
        rds = c.image.attrs['RepoDigests']
        return DockerImageDefInstance([tuple(rd.split("@")) for rd in rds])


@dataclass
class DockerEnvDefInstance(DockerAttributeInstance):
    env: t.Mapping[str, str]

    def to_run_arg(self) -> t.Any:
        return self.env

    def is_satisfied_by(self, o: 'DockerEnvDefInstance'):
        desired = self
        actual = o
        return all((k in actual.env and actual.env[k] == str(v)) for k, v in desired.env.items())


class DockerEnvDef(DockerAttribute):

    def run_arg_name(self) -> str:
        return "environment"

    def from_necronomicon(self, dc: DockerContainer) -> 'DockerAttributeInstance':
        return DockerEnvDefInstance(dc.env)

    def from_container(self, c: Container) -> 'DockerAttributeInstance':
        found_env_list: t.List[str] = c.attrs['Config']['Env']
        found_env = {l[0]: l[1] for l in (l.split("=", 1) for l in found_env_list)}
        return DockerEnvDefInstance(found_env)


class DockerContainerNameDef(DockerAttribute):

    def run_arg_name(self) -> str:
        return "name"

    def from_necronomicon(self, dc: DockerContainer) -> 'DockerAttributeInstance':
        return SimpleDockerAttributeInstance(dc.name)

    def from_container(self, c: Container) -> 'DockerAttributeInstance':
        return SimpleDockerAttributeInstance(c.name)


@dataclass
class DockerPortsDefInstance(DockerAttributeInstance):
    ports: t.List[PortEntry]

    def to_run_arg(self) -> t.Dict[t.Union[str, int], t.Union[int, t.Tuple[str, int]]]:
        return {pe.container.as_run_arg_key(): [hv.as_run_arg_value() for hv in pe.host] for pe in self.ports}

    def is_satisfied_by(self, actual: 'DockerPortsDefInstance'):
        return sorted(self.ports) == sorted(actual.ports)


class DockerPortsDef(DockerAttribute):
    def run_arg_name(self) -> str:
        return "ports"

    def from_necronomicon(self, dc: DockerContainer) -> 'DockerAttributeInstance':
        return DockerPortsDefInstance(dc.ports)

    def from_container(self, c: Container) -> 'DockerAttributeInstance':
        """
        Example: {
            '33200/tcp': [{'HostIp': '0.0.0.0', 'HostPort': '33200'}],
            '53/tcp': [{'HostIp': '192.168.1.103', 'HostPort': '53'}],
            '53/udp': [{'HostIp': '192.168.1.103', 'HostPort': '53'}]
        }
        """

        ret = []
        for container_port, host_ports in c.ports.items():
            ret.append(PortEntry(
                PortString.from_str(container_port, allow_ip=False),
                sorted([PortString.from_tuple(ipaddress.ip_address(hp['HostIp']) if hp['HostIp'] != '0.0.0.0' else None, int(hp['HostPort']), None, allow_proto=False) for hp in host_ports if ipaddress.ip_address(hp['HostIp']).version == 4]) if host_ports else [],
            ))

        return DockerPortsDefInstance(ret)


@dataclass
class DockerVolumesDefInstance(DockerAttributeInstance):
    volumes: t.Mapping[str, str]

    def to_run_arg(self) -> t.Any:
        ret = {}
        for k, v in self.volumes.items():
            mode_data = re.search(r"\+(?P<mode>r[ow])$", v)
            if mode_data:
                tmp = {"bind": v[:-3], "mode": mode_data.group("mode").strip()}
                if tmp["bind"].endswith("/"):
                    tmp["bind"] = tmp["bind"][:-1]
                ret[k] = tmp
            else:
                ret[k] = {"bind": v[:-1] if v.endswith("/") else v, "mode": "rw"}

        return ret

    def is_satisfied_by(self, other: 'DockerVolumesDefInstance'):
        return self.volumes == other.volumes


class DockerVolumesDef(DockerAttribute):

    def run_arg_name(self) -> str:
        return "volumes"

    def from_necronomicon(self, dc: DockerContainer) -> 'DockerVolumesDefInstance':
        ret = {}
        for k, v in dc.volumes.items():
            if not re.search(r"\+(?P<mode>r[ow])$", v):
                v = f"{v}+rw"
            if v[-4] == "/":
                v = v[:-4] + v[-3:]
            ret[k] = v
        return DockerVolumesDefInstance(ret)

    def from_container(self, c: Container) -> 'DockerVolumesDefInstance':
        found_mounts_list: t.List[t.Dict[str, t.Any]] = c.attrs['Mounts']
        found_mounts_volume = {m['Name']: {"bind": m['Destination'], "mode": m["Mode"]} for m in found_mounts_list if
                               m['Type'] == 'volume'}
        found_mounts_bind = {m['Source']: {"bind": m['Destination'], "mode": m["Mode"]} for m in found_mounts_list if
                             m['Type'] == 'bind'}
        found_mounts = {}
        for k, v in found_mounts_volume.items():
            found_mounts[k] = f"{v['bind']}+{v['mode']}"

        for k, v in found_mounts_bind.items():
            found_mounts[k] = f"{v['bind']}+{v['mode']}"

        return DockerVolumesDefInstance(found_mounts)


@dataclass
class DockerCommandDefInstance(DockerAttributeInstance):
    command: t.Union[t.List[str], None]
    image: t.Union[Image, None]

    def to_run_arg(self) -> t.Any:
        return self.command

    def is_satisfied_by(self, actual: 'DockerCommandDefInstance'):
        if self.command:
            return self.command == actual.command
        else:
            return actual.image.attrs['Config']['Cmd'] == actual.command


class DockerCommandDef(DockerAttribute):
    def run_arg_name(self) -> str:
        return "command"

    def from_necronomicon(self, dc: DockerContainer) -> 'DockerAttributeInstance':
        return DockerCommandDefInstance(dc.command.split(" ") if dc.command else None, None)

    def from_container(self, c: Container) -> 'DockerAttributeInstance':
        return DockerCommandDefInstance(c.attrs["Config"]["Cmd"], c.image)


class DockerSysctlDef(DockerAttribute):
    sysctls: t.Mapping[str, str]

    def run_arg_name(self) -> str:
        return "sysctls"

    def from_necronomicon(self, dc: DockerContainer) -> 'DockerAttributeInstance':
        return SimpleDockerAttributeInstance(dc.sysctls)

    def from_container(self, c: Container) -> 'DockerAttributeInstance':
        if 'Sysctls' in c.attrs['HostConfig']:
            return SimpleDockerAttributeInstance(c.attrs['HostConfig']['Sysctls'])
        else:
            return SimpleDockerAttributeInstance({})


class DockerCapabilitiesDef(DockerAttribute):

    def run_arg_name(self) -> str:
        return "cap_add"

    def from_necronomicon(self, dc: DockerContainer) -> 'DockerAttributeInstance':
        return SimpleDockerAttributeInstance(dc.capabilities)

    def from_container(self, c: Container) -> 'DockerAttributeInstance':
        theirs = c.attrs['HostConfig']['CapAdd']
        if theirs is None:
            theirs = []
        return SimpleDockerAttributeInstance(theirs)


class DockerDetachDef(DockerAttribute):
    def run_arg_name(self) -> str:
        return "detach"

    def from_necronomicon(self, dc: DockerContainer) -> 'DockerAttributeInstance':
        return SimpleDockerAttributeInstance(True)

    def from_container(self, c: Container) -> 'DockerAttributeInstance':
        return SimpleDockerAttributeInstance(True)


class DockerLoggingDef(DockerAttribute):
    def run_arg_name(self) -> str:
        return "log_config"

    def from_necronomicon(self, dc: DockerContainer) -> 'DockerAttributeInstance':
        return SimpleDockerAttributeInstance(LogConfig(type=LogConfig.types.JOURNALD))

    def from_container(self, c: Container) -> 'DockerAttributeInstance':
        return SimpleDockerAttributeInstance(c.attrs['HostConfig']['LogConfig'])


class DockerRestartPolicyDef(DockerAttribute):
    def run_arg_name(self) -> str:
        return "restart_policy"

    def from_necronomicon(self, dc: DockerContainer) -> 'DockerAttributeInstance':
        return SimpleDockerAttributeInstance({'Name': "always"})

    def from_container(self, c: Container) -> 'DockerAttributeInstance':
        found = c.attrs['HostConfig']['RestartPolicy']
        return SimpleDockerAttributeInstance({'Name': found['Name']})


class DockerNameDef(DockerAttribute):
    def run_arg_name(self) -> str:
        return "name"

    def from_necronomicon(self, dc: DockerContainer) -> 'DockerAttributeInstance':
        return SimpleDockerAttributeInstance(dc.name)

    def from_container(self, c: Container) -> 'DockerAttributeInstance':
        return SimpleDockerAttributeInstance(c.name)


class DockerNetworkingModeDef(DockerAttribute):
    def run_arg_name(self) -> str:
        return "network_mode"

    def from_necronomicon(self, dc: DockerContainer) -> 'DockerAttributeInstance':
        return SimpleDockerAttributeInstance(dc.network_mode)

    def from_container(self, c: Container) -> 'DockerAttributeInstance':
        found = c.attrs['HostConfig']['NetworkMode']
        # if found == "default":
        #     found = "bridge"
        return SimpleDockerAttributeInstance(found)


class DockerNetworkingModeDefInstance(DockerAttributeInstance):
    def __init__(self, network_mode_name):
        self.network_mode_name = network_mode_name

    def to_run_arg(self) -> t.Any:
        if self.network_mode_name == "default":
            return None
        else:
            return self.network_mode_name

    def is_satisfied_by(self, other: 'DockerNetworkingModeDefInstance'):
        return self.network_mode_name == other.network_mode_name

    def __repr__(self) -> str:
        return repr(self.network_mode_name)


def diff_container(c: Container, dc: DockerContainer) -> t.List[t.Tuple[str, t.Any, t.Any]]:
    diffs = []
    for da in SUPPORTED_DOCKER_ATTRS:
        desired = da.from_necronomicon(dc)
        found = da.from_container(c)

        if not desired.is_satisfied_by(found):
            diffs.append((da.run_arg_name(), desired, found))
    return diffs


SUPPORTED_DOCKER_ATTRS: t.List['DockerAttribute'] = [
    DockerImageDef(),
    DockerEnvDef(),
    DockerContainerNameDef(),
    DockerPortsDef(),
    DockerVolumesDef(),
    DockerCommandDef(),
    DockerSysctlDef(),
    DockerCapabilitiesDef(),
    DockerDetachDef(),
    DockerLoggingDef(),
    DockerRestartPolicyDef(),
    DockerNameDef(),
    DockerNetworkingModeDef(),
]



import logging
import typing as t
from dataclasses import dataclass
from ipaddress import ip_address, IPv4Address

import yaml
from paramiko.client import SSHClient
from yog.model_utils import HostPathStr, parse_hostpath

from yog.ssh_utils import check_stdout


def loads(ident: str, necronomicon: str) -> 'Necronomicon':
    return load(ident, yaml.safe_load(necronomicon))


def loadfile(ident: str, path: str) -> 'Necronomicon':
    with open(path) as nin:
        return loads(ident, nin.read())


def load(ident: str, parsed_necronomicon) -> 'Necronomicon':
    if parsed_necronomicon is None:
        return Necronomicon(ident, NeededTunnelsSection([]), DockerSection([]), CronSection([]), FileSection([]))

    if 'files' in parsed_necronomicon:
        fs = FileSection([File(
            e['src'],
            e['dest'],
            e['hupcmd'] if 'hupcmd' in e else None,
            e['root'] if 'root' in e else False,
        ) for e in parsed_necronomicon['files']])
    else:
        fs = FileSection([])

    if 'docker' in parsed_necronomicon:
        ds = DockerSection([DockerContainer(
            e['image'],
            e['name'],
            e['fingerprint'],
            e['volumes'] if 'volumes' in e else {},
            [PortEntry.from_yaml(pe) for pe in (e['ports'] if 'ports' in e else [])],
            e['env'] if 'env' in e else {},
            e['command'] if 'command' in e else None,
            e['capabilities'] if 'capabilities' in e else [],
            {str(k): str(v) for k, v in e['sysctls'].items()} if 'sysctls' in e else {},
            e['network_mode'] if 'network_mode' in e else "default",
        ) for e in parsed_necronomicon['docker']])
    else:
        ds = DockerSection([])

    if 'cron' in parsed_necronomicon:
        cs = CronSection([CronJob(e['expr'], e['command'], e['user'] if 'user' in e else 'root') for e in parsed_necronomicon['cron']])
    else:
        cs = CronSection([])

    if 'needs_tunnels' in parsed_necronomicon:
        tunnels = NeededTunnelsSection([
            NeededTunnel(
                tun['host'],
                int(tun['target_port']),
                int(tun['local_port']),
            ) for tun in parsed_necronomicon['needs_tunnels']
        ])
    else:
        tunnels = NeededTunnelsSection([])

    if 'pki' in parsed_necronomicon:
        pki = PKI([CertEntry.from_parsed(e) for e in parsed_necronomicon['pki']['certs']] if 'certs' in parsed_necronomicon['pki'] else [],
                  # parsed_necronomicon['pki']['authorities'] if 'authorities' in parsed_necronomicon['pki'] else []
                  )
    else:
        pki = PKI([])

    return Necronomicon(ident, tunnels, ds, cs, fs, pki)


class Necronomicon(t.NamedTuple):
    ident: str
    tunnels: 'NeededTunnelsSection'
    docker: 'DockerSection'
    cron: 'CronSection'
    files: 'FileSection'
    pki: 'PKI'

    def inflate(self, host: str, ssh: SSHClient) -> 'Necronomicon':
        inflated_containers = []
        for desired_container in self.docker.containers:
            inflated_desired_container_env = {}
            for name, value in desired_container.env.items():
                value = str(value)
                if value.startswith("yogreadfile:"):
                    try:
                        inflated_desired_container_env[name] = "\n".join(
                            check_stdout(ssh, f"sudo cat {value[len('yogreadfile:'):]}")).strip()
                    except RuntimeError as err:
                        logging.error(f"Error processing yogreadfile: {value}", exc_info=err)
                        raise RuntimeError(f"Error accessing file: {value[len('yogreadfile:'):]}")
                else:
                    inflated_desired_container_env[name] = value
            inflated_containers.append(DockerContainer(
                desired_container.image,
                desired_container.name,
                desired_container.fingerprint,
                desired_container.volumes,
                desired_container.ports,
                inflated_desired_container_env,
                desired_container.command,
                desired_container.capabilities,
                desired_container.sysctls,
                desired_container.network_mode,
            ))

        return Necronomicon(
            self.ident,
            self.tunnels,
            DockerSection(inflated_containers),
            self.cron,
            self.files,
            self.pki,
        )


class DockerSection(t.NamedTuple):
    containers: t.List['DockerContainer']


class DockerContainer(t.NamedTuple):
    image: str
    name: str
    fingerprint: str
    volumes: t.Mapping[str, str]
    ports: t.List['PortEntry']
    env: t.Mapping[str, str]
    command: t.Optional[str]
    capabilities: t.List[str]
    sysctls: t.Mapping[str, str]
    network_mode: str


class PortEntry(t.NamedTuple):
    container: 'PortString'
    host: t.List['PortString']

    @staticmethod
    def from_yaml(y) -> 'PortEntry':
        return PortEntry(
            PortString.from_str(y["container"], allow_ip=False),
            sorted([PortString.from_str(s, allow_proto=False) for s in y["host"]]),
        )

    def __str__(self):
        host_part = ",".join([str(ps) for ps in self.host])
        return f"{self.container}->{host_part}"

    def __repr__(self) -> str:
        return str(self)

    def __lt__(self, other):
        return str(self) < str(other)


@dataclass
class PortString:
    ip: t.Union[IPv4Address, None]
    port: int
    proto: t.Union[str, None]

    def as_run_arg_key(self) -> t.Union[str, int]:
        if self.proto:
            return f"{self.port}/{self.proto}"
        else:
            return self.port

    def as_run_arg_value(self) -> t.Union[int, t.Tuple[str, int]]:
        if self.ip:
            return str(self.ip), self.port
        else:
            return self.port

    def __str__(self) -> str:
        return f"{str(self.ip) if self.ip else '0.0.0.0'}:{self.port}/{self.proto if self.proto else 'any'}"

    def __eq__(self, o: object) -> bool:
        return str(self) == str(o)

    def __lt__(self, other):
        return str(self) < str(other)

    @staticmethod
    def _normalize(ip: t.Union[IPv4Address, None], port: int, proto: t.Union[str, None], allow_proto=True, allow_ip=True):
        if not allow_ip and ip:
            raise ValueError(f"IP is not allowed (is meaningless) for port expr {ip},{port},{proto}")
        if not allow_proto and proto:
            raise ValueError(F"Protocol is not allowed (is meaningless) for port expr {ip},{port},{proto}")

        if allow_proto and not proto:
            proto = 'tcp'
        if allow_ip and not ip:
            ip = ip_address('0.0.0.0')

        return ip, port, proto

    @staticmethod
    def from_tuple(ip: t.Union[IPv4Address, None], port: int, proto: t.Union[str, None], allow_proto=True, allow_ip=True):
        ip, port, proto = PortString._normalize(ip, port, proto, allow_proto, allow_ip)
        return PortString(ip, port, proto)

    @staticmethod
    def from_str(s: str, allow_proto=True, allow_ip=True) -> 'PortString':
        if isinstance(s, int):
            s = str(s)

        if "/" in s:
            ipport, proto = s.split("/")
        else:
            ipport = s
            proto = None

        if ":" in ipport:
            ip, port = ipport.split(":")
            port = int(port)
        else:
            port = int(ipport)
            ip = None

        ip, port, proto = PortString._normalize(ip, port, proto, allow_proto, allow_ip)
        return PortString(ip_address(ip) if ip else None, port, proto)


class CronSection(t.NamedTuple):
    crons: t.List['CronJob']


class CronJob(t.NamedTuple):
    expr: str
    command: str
    user: str


class FileSection(t.NamedTuple):
    files: t.List['File']


class File(t.NamedTuple):
    src: str
    dest: str
    hupcmd: str
    root: bool


class NeededTunnel(t.NamedTuple):
    host: str
    target_port: int
    local_port: int


class NeededTunnelsSection(t.NamedTuple):
    tunnels: t.List[NeededTunnel]


class CertEntry(t.NamedTuple):
    storage: str
    validity_period: str
    refresh_at_period: str
    names: t.List[str]
    authority: str
    hupcmd: t.Optional[str]
    chmod: t.Optional[str]
    chown: t.Optional[str]

    @staticmethod
    def from_parsed(o: t.Any):
        return CertEntry(
            o['storage'],
            str(o['validity_period']),
            str(o['refresh_at_period']),
            o['names'],
            o['authority'],
            o['hupcmd'] if 'hupcmd' in o else None,
            o['chmod'] if 'chmod' in o else "rw-r--r--",
            o['chown'] if 'chown' in o else "root:root",
        )


class PKI(t.NamedTuple):
    certs: t.List[CertEntry]




import logging
import unittest
from unittest.mock import MagicMock

import yaml
from docker.models.containers import Container
from docker.models.images import Image

from yog.host.docker_attrs import SUPPORTED_DOCKER_ATTRS
from yog.host.necronomicon import DockerContainer, load
from yog.res import get_resource


class TestDockerAttrs(unittest.TestCase):

    def test_docker_container_equals(self):
        n = load("dockertest", yaml.safe_load(get_resource("docker_test.yml")))

        c: MagicMock = MagicMock(spec=Container)

        c.attrs = {'Id': '73d4999164755197142a155133c022baec3817ad91071741927bcc2f384c7f0c', 'Created': '2023-01-30T03:40:36.570012116Z', 'Path': 'docker-entrypoint.sh', 'Args': ['/usr/bin/dumb-init', 'node', 'server.js'], 'State': {'Status': 'running', 'Running': True, 'Paused': False, 'Restarting': False, 'OOMKilled': False, 'Dead': False, 'Pid': 1282, 'ExitCode': 0, 'Error': '', 'StartedAt': '2023-01-30T05:12:09.960486407Z', 'FinishedAt': '2023-01-30T04:38:46.01411325Z'}, 'Image': 'sha256:c90b263d8a383ac55905e4edbe1513ba87b1a479bcaa8bb471412137b9541751', 'ResolvConfPath': '/var/lib/docker/containers/73d4999164755197142a155133c022baec3817ad91071741927bcc2f384c7f0c/resolv.conf', 'HostnamePath': '/var/lib/docker/containers/73d4999164755197142a155133c022baec3817ad91071741927bcc2f384c7f0c/hostname', 'HostsPath': '/var/lib/docker/containers/73d4999164755197142a155133c022baec3817ad91071741927bcc2f384c7f0c/hosts', 'LogPath': '', 'Name': '/wg-in-test', 'RestartCount': 0, 'Driver': 'overlay2', 'Platform': 'linux', 'MountLabel': '', 'ProcessLabel': '', 'AppArmorProfile': 'docker-default', 'ExecIDs': None, 'HostConfig': {'Binds': ['/srv/wg-in/config:/etc/wireguard:rw'], 'ContainerIDFile': '', 'LogConfig': {'Type': 'journald', 'Config': {}}, 'NetworkMode': 'default', 'PortBindings': {'51820/udp': [{'HostIp': '0.0.0.0', 'HostPort': '51820'}], '51821/tcp': [{'HostIp': '0.0.0.0', 'HostPort': '51821'}]}, 'RestartPolicy': {'Name': 'always', 'MaximumRetryCount': 0}, 'AutoRemove': False, 'VolumeDriver': '', 'VolumesFrom': None, 'CapAdd': ['NET_ADMIN', 'SYS_MODULE'], 'CapDrop': None, 'CgroupnsMode': 'host', 'Dns': [], 'DnsOptions': [], 'DnsSearch': [], 'ExtraHosts': None, 'GroupAdd': None, 'IpcMode': 'private', 'Cgroup': '', 'Links': None, 'OomScoreAdj': 0, 'PidMode': '', 'Privileged': False, 'PublishAllPorts': False, 'ReadonlyRootfs': False, 'SecurityOpt': None, 'UTSMode': '', 'UsernsMode': '', 'ShmSize': 67108864, 'Sysctls': {'net.ipv4.conf.all.forwarding': '1', 'net.ipv4.conf.all.proxy_arp': '1', 'net.ipv4.conf.all.src_valid_mark': '1', 'net.ipv4.ip_forward': '1'}, 'Runtime': 'runc', 'ConsoleSize': [0, 0], 'Isolation': '', 'CpuShares': 0, 'Memory': 0, 'NanoCpus': 0, 'CgroupParent': '', 'BlkioWeight': 0, 'BlkioWeightDevice': None, 'BlkioDeviceReadBps': None, 'BlkioDeviceWriteBps': None, 'BlkioDeviceReadIOps': None, 'BlkioDeviceWriteIOps': None, 'CpuPeriod': 0, 'CpuQuota': 0, 'CpuRealtimePeriod': 0, 'CpuRealtimeRuntime': 0, 'CpusetCpus': '', 'CpusetMems': '', 'Devices': None, 'DeviceCgroupRules': None, 'DeviceRequests': None, 'KernelMemory': 0, 'KernelMemoryTCP': 0, 'MemoryReservation': 0, 'MemorySwap': 0, 'MemorySwappiness': None, 'OomKillDisable': False, 'PidsLimit': None, 'Ulimits': None, 'CpuCount': 0, 'CpuPercent': 0, 'IOMaximumIOps': 0, 'IOMaximumBandwidth': 0, 'MaskedPaths': ['/proc/asound', '/proc/acpi', '/proc/kcore', '/proc/keys', '/proc/latency_stats', '/proc/timer_list', '/proc/timer_stats', '/proc/sched_debug', '/proc/scsi', '/sys/firmware'], 'ReadonlyPaths': ['/proc/bus', '/proc/fs', '/proc/irq', '/proc/sys', '/proc/sysrq-trigger']}, 'GraphDriver': {'Data': {'LowerDir': '/var/lib/docker/overlay2/d85f7cf20def119cbaad54078be5cdfa882284bc5be2ddb52330990486e55641-init/diff:/var/lib/docker/overlay2/8c83b36346afb47016a59a8ecd18c1e0e1ecd8cf85b6f3f4f07c051d864e9f93/diff:/var/lib/docker/overlay2/78c7097986814469e6dd0604d0eb5ae92715d5d3da1cbd2525abf1c78b23003a/diff:/var/lib/docker/overlay2/6e0e74616ea70f1df32bfd1779c10dbe69f8d4f4a643961113ba77f3f9a30f88/diff:/var/lib/docker/overlay2/d1aa53c2061daadd88130940e1e81fb0b409f6e3112d91950a5f27e37db25b2d/diff:/var/lib/docker/overlay2/2dd7adcad9281d851b601fa42690bd6263e51a0311768055878c180718aca963/diff:/var/lib/docker/overlay2/cc93b6479b36f400e8b77afef5bbb01d471f3513a78dfcf2755b4e67bb4b12b4/diff:/var/lib/docker/overlay2/6d1efc10653abb13b983e3a908bb8ceec3b1a83637e8918c98e2996f967558f2/diff:/var/lib/docker/overlay2/d4977d659e900e6a14f648a50b3b0441a27b110142cc48b9a33c24411ec6b479/diff:/var/lib/docker/overlay2/5c9bb6d6187ebc71d81f92d32ae2e56e06f3cc4fd2f0d40b9b5809249d8d1c6e/diff', 'MergedDir': '/var/lib/docker/overlay2/d85f7cf20def119cbaad54078be5cdfa882284bc5be2ddb52330990486e55641/merged', 'UpperDir': '/var/lib/docker/overlay2/d85f7cf20def119cbaad54078be5cdfa882284bc5be2ddb52330990486e55641/diff', 'WorkDir': '/var/lib/docker/overlay2/d85f7cf20def119cbaad54078be5cdfa882284bc5be2ddb52330990486e55641/work'}, 'Name': 'overlay2'}, 'Mounts': [{'Type': 'bind', 'Source': '/srv/wg-in/config', 'Destination': '/etc/wireguard', 'Mode': 'rw', 'RW': True, 'Propagation': 'rprivate'}], 'Config': {'Hostname': '73d499916475', 'Domainname': '', 'User': '', 'AttachStdin': False, 'AttachStdout': False, 'AttachStderr': False, 'ExposedPorts': {'51820/udp': {}, '51821/tcp': {}}, 'Tty': False, 'OpenStdin': False, 'StdinOnce': False, 'Env': ['WG_HOST=ofn.wg.josh.cafe', 'WG_PORT=48888', 'PASSWORD=testpassword', 'WG_DEFAULT_DNS=192.168.1.103, 192.168.1.189', 'WG_DEFAULT_ADDRESS=192.168.100.x', 'PATH=/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin', 'NODE_VERSION=14.18.1', 'YARN_VERSION=1.22.15', 'DEBUG=Server,WireGuard'], 'Cmd': ['/usr/bin/dumb-init', 'node', 'server.js'], 'Image': 'weejewel/wg-easy@sha256:79288cac4782dd9c37c2fc20d26eb663d2df9fa1d3a09b278d5c76411310ec0c', 'Volumes': {'/etc/wireguard': {}}, 'WorkingDir': '/app', 'Entrypoint': ['docker-entrypoint.sh'], 'OnBuild': None, 'Labels': {}}, 'NetworkSettings': {'Bridge': '', 'SandboxID': 'f38a10fd4125fe1f8115efeb2dd9366c29baf9614bd3e8aebefe998b11e2a1fe', 'HairpinMode': False, 'LinkLocalIPv6Address': '', 'LinkLocalIPv6PrefixLen': 0, 'Ports': {'51820/udp': [{'HostIp': '0.0.0.0', 'HostPort': '51820'}], '51821/tcp': [{'HostIp': '0.0.0.0', 'HostPort': '51821'}]}, 'SandboxKey': '/var/run/docker/netns/f38a10fd4125', 'SecondaryIPAddresses': None, 'SecondaryIPv6Addresses': None, 'EndpointID': 'bb4fea7c4159ad4f5de82a8bc1fc1415d23a916e0d1d4a8df8487d4b170256ff', 'Gateway': '172.17.0.1', 'GlobalIPv6Address': '', 'GlobalIPv6PrefixLen': 0, 'IPAddress': '172.17.0.2', 'IPPrefixLen': 16, 'IPv6Gateway': '', 'MacAddress': '02:42:ac:11:00:02', 'Networks': {'bridge': {'IPAMConfig': None, 'Links': None, 'Aliases': None, 'NetworkID': 'dfbd02fb880c4b9cfb11191c395703e2fbd439d5d7a0953ba9dbd933713a1476', 'EndpointID': 'bb4fea7c4159ad4f5de82a8bc1fc1415d23a916e0d1d4a8df8487d4b170256ff', 'Gateway': '172.17.0.1', 'IPAddress': '172.17.0.2', 'IPPrefixLen': 16, 'IPv6Gateway': '', 'GlobalIPv6Address': '', 'GlobalIPv6PrefixLen': 0, 'MacAddress': '02:42:ac:11:00:02', 'DriverOpts': None}}}}
        c.ports = {'51820/udp': [{'HostIp': '0.0.0.0', 'HostPort': '51820'}], '51821/tcp': [{'HostIp': '0.0.0.0', 'HostPort': '51821'}]}
        c.name = 'wg-in-test'

        img: MagicMock = MagicMock(spec=Image)
        img.attrs = {'Id': 'sha256:c90b263d8a383ac55905e4edbe1513ba87b1a479bcaa8bb471412137b9541751', 'RepoTags': [], 'RepoDigests': ['weejewel/wg-easy@sha256:79288cac4782dd9c37c2fc20d26eb663d2df9fa1d3a09b278d5c76411310ec0c'], 'Parent': '', 'Comment': 'buildkit.dockerfile.v0', 'Created': '2022-06-26T09:30:56.818800131Z', 'Container': '', 'ContainerConfig': {'Hostname': '', 'Domainname': '', 'User': '', 'AttachStdin': False, 'AttachStdout': False, 'AttachStderr': False, 'Tty': False, 'OpenStdin': False, 'StdinOnce': False, 'Env': None, 'Cmd': None, 'Image': '', 'Volumes': None, 'WorkingDir': '', 'Entrypoint': None, 'OnBuild': None, 'Labels': None}, 'DockerVersion': '', 'Author': '', 'Config': {'Hostname': '', 'Domainname': '', 'User': '', 'AttachStdin': False, 'AttachStdout': False, 'AttachStderr': False, 'ExposedPorts': {'51820/udp': {}, '51821/tcp': {}}, 'Tty': False, 'OpenStdin': False, 'StdinOnce': False, 'Env': ['PATH=/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin', 'NODE_VERSION=14.18.1', 'YARN_VERSION=1.22.15', 'DEBUG=Server,WireGuard'], 'Cmd': ['/usr/bin/dumb-init', 'node', 'server.js'], 'ArgsEscaped': True, 'Image': '', 'Volumes': None, 'WorkingDir': '/app', 'Entrypoint': ['docker-entrypoint.sh'], 'OnBuild': None, 'Labels': None}, 'Architecture': 'amd64', 'Os': 'linux', 'Size': 141586777, 'VirtualSize': 141586777, 'GraphDriver': {'Data': {'LowerDir': '/var/lib/docker/overlay2/78c7097986814469e6dd0604d0eb5ae92715d5d3da1cbd2525abf1c78b23003a/diff:/var/lib/docker/overlay2/6e0e74616ea70f1df32bfd1779c10dbe69f8d4f4a643961113ba77f3f9a30f88/diff:/var/lib/docker/overlay2/d1aa53c2061daadd88130940e1e81fb0b409f6e3112d91950a5f27e37db25b2d/diff:/var/lib/docker/overlay2/2dd7adcad9281d851b601fa42690bd6263e51a0311768055878c180718aca963/diff:/var/lib/docker/overlay2/cc93b6479b36f400e8b77afef5bbb01d471f3513a78dfcf2755b4e67bb4b12b4/diff:/var/lib/docker/overlay2/6d1efc10653abb13b983e3a908bb8ceec3b1a83637e8918c98e2996f967558f2/diff:/var/lib/docker/overlay2/d4977d659e900e6a14f648a50b3b0441a27b110142cc48b9a33c24411ec6b479/diff:/var/lib/docker/overlay2/5c9bb6d6187ebc71d81f92d32ae2e56e06f3cc4fd2f0d40b9b5809249d8d1c6e/diff', 'MergedDir': '/var/lib/docker/overlay2/8c83b36346afb47016a59a8ecd18c1e0e1ecd8cf85b6f3f4f07c051d864e9f93/merged', 'UpperDir': '/var/lib/docker/overlay2/8c83b36346afb47016a59a8ecd18c1e0e1ecd8cf85b6f3f4f07c051d864e9f93/diff', 'WorkDir': '/var/lib/docker/overlay2/8c83b36346afb47016a59a8ecd18c1e0e1ecd8cf85b6f3f4f07c051d864e9f93/work'}, 'Name': 'overlay2'}, 'RootFS': {'Type': 'layers', 'Layers': ['sha256:39982b2a789afc156fff00c707d0ff1c6ab4af8f1666a8df4787714059ce24e7', 'sha256:a303372b2caae8beec770cc39e25d3f8ef61d39bdd6ddfb35e408fc402b9cf96', 'sha256:b3031b5001d5acbb8e13c1ec54708914a524b6fff54c49a9e8968e9c9672f249', 'sha256:e763407a02c7e01cd850dd14d8332b6ea1fb8e5f3c016df96bf305ef0ad47826', 'sha256:159322f2828b0337b56c60a350c02bdc6444575fc50ab3c8684499db58e543b2', 'sha256:b7fb9f18b6d36619ec0c0a413320c02aa929d92628213d33ca3608c4e7d49800', 'sha256:83ef97c139857b0f3460a1fdbba93e527012de34b04899d101d9c5131f7e4b73', 'sha256:a8331a10795a20ba653f538f78fd41d1717dbde3a24a24db9cebe6402dde2a37', 'sha256:5f70bf18a086007016e948b04aed3b82103a36bea41755b6cddfaf10ace3c6ef']}, 'Metadata': {'LastTagTime': '0001-01-01T00:00:00Z'}}

        c.image = img

        dc: DockerContainer = n.docker.containers[0]

        for da in SUPPORTED_DOCKER_ATTRS:
            ours = da.from_necronomicon(dc)
            theirs = da.from_container(c)
            logging.info(f"{da.run_arg_name()}: o({ours}) t({theirs})")
            self.assertTrue(ours.is_satisfied_by(theirs), f"{da.run_arg_name()} should match but desired={ours} and found is {theirs}")


if __name__ == '__main__':
    unittest.main()
